#pragma once

#include "../../DNACommon/DNACommon.hpp"
#include "IScriptObject.hpp"
#include "Parameters.hpp"

namespace DataSpec::DNAMP1 {
struct NewIntroBoss : IScriptObject {
  AT_DECL_DNA_YAMLV
  String<-1> name;
  Value<atVec3f> location;
  Value<atVec3f> orientation;
  Value<atVec3f> scale;
  PatternedInfo patternedInfo;
  ActorParameters actorParameters;
  Value<float> minTurnAngle;
  UniqueID32 weaponDesc;
  DamageInfo damageInfo;
  UniqueID32 beamContactFxId;
  UniqueID32 beamPulseFxId;
  UniqueID32 beamTextureId;
  UniqueID32 beamGlowTextureId;

  void addCMDLRigPairs(PAKRouter<PAKBridge>& pakRouter, CharacterAssociations<UniqueID32>& charAssoc) const override {
    actorParameters.addCMDLRigPairs(pakRouter, charAssoc, patternedInfo.animationParameters);
  }

  void nameIDs(PAKRouter<PAKBridge>& pakRouter) const override {
    if (beamContactFxId.isValid()) {
      PAK::Entry* ent = (PAK::Entry*)pakRouter.lookupEntry(beamContactFxId);
      ent->name = name + "_beamContactFxId";
    }
    if (beamPulseFxId.isValid()) {
      PAK::Entry* ent = (PAK::Entry*)pakRouter.lookupEntry(beamPulseFxId);
      ent->name = name + "_beamPulseFxId";
    }
    if (beamTextureId.isValid()) {
      PAK::Entry* ent = (PAK::Entry*)pakRouter.lookupEntry(beamTextureId);
      ent->name = name + "_beamTextureId";
    }
    if (beamGlowTextureId.isValid()) {
      PAK::Entry* ent = (PAK::Entry*)pakRouter.lookupEntry(beamGlowTextureId);
      ent->name = name + "_beamGlowTextureId";
    }
    patternedInfo.nameIDs(pakRouter, name + "_patterned");
    actorParameters.nameIDs(pakRouter, name + "_actp");
  }

  void gatherDependencies(std::vector<hecl::ProjectPath>& pathsOut,
                          std::vector<hecl::ProjectPath>& lazyOut) const override {
    g_curSpec->flattenDependencies(weaponDesc, pathsOut);
    g_curSpec->flattenDependencies(beamContactFxId, pathsOut);
    g_curSpec->flattenDependencies(beamPulseFxId, pathsOut);
    g_curSpec->flattenDependencies(beamTextureId, pathsOut);
    g_curSpec->flattenDependencies(beamGlowTextureId, pathsOut);
    patternedInfo.depIDs(pathsOut);
    actorParameters.depIDs(pathsOut, lazyOut);
  }

  void gatherScans(std::vector<Scan>& scansOut) const override { actorParameters.scanIDs(scansOut); }
};
} // namespace DataSpec::DNAMP1
